# coding=utf-8
# Copyright 2020 The Private Medians Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Medians mechanism comparison experiments for ICML 2020 submission.

Methods: smooth sensitivity mechanism and exponential mechanism.
Data generated from: N(0,1), U(0,1), and Beta(0.5, 0.5).

Prints statistics and saves plots to /tmp/private_medians.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import functools
import os
import random
import time

import matplotlib.pyplot as plt
import numpy as np

import private_medians

# Fix random seed for reproducibility.
_RANDOM_SEED = 1

# Set to true to print exact stats.
_VERBOSE = True

# Size of dataset.
_DATASET_SIZE = 1000

# Patience of test-runner :).
_NUM_DATASETS = 100
_NUM_TRIALS_PER_DATASET = 100

# Privacy parameter.
_EPSILONS = [0.1, 0.5, 1.0, 1.5, 2.0]

# Delta is for smooth sensitivity only.
_DELTA0 = 0
_DELTA_SMALL = 1.0 / _DATASET_SIZE
_DELTA_LARGE = 0.9


def run_medians_eval(data_gen_fun, n, min_value, max_value, num_datasets,
                     num_trials_per_dataset, methods, names, verbose):
  """Evaluates private medians algorithms on datasets generated by data_gen_fun.

  Args:
    data_gen_fun : Function that takes as input a size and generates a dataset
      of that size.
    n: Desired dataset size.
    min_value: Min value allowed in a dataset.  Values lower than this will be
      incorporated in the computation of the true median, but will be truncated
      when calling private medians algorithms.
    max_value: Max value allowed in a dataset.  Values above this will be
      incorporated in the computation of the true median, but will be truncated
      when calling private medians algorithms.
    num_datasets: Number of datasets to generate.
    num_trials_per_dataset: Number of times to run each private algorithm per
      dataset.
    methods: List of differentially-private medians mechanisms to try.  Each
      method in this list is a function that takes as input a sorted list and a
      number of estimates and returns a list of private median estimates.
    names: List of names for each of the methods to annotate the results.
    verbose: Boolean to indicate whether to print a verbose output.

  Returns:
    A list of average error across all datasets for all different methods and
    a list of standard deviations.
  """
  num_methods = len(methods)
  num_methods_range = range(num_methods)

  diffs = [[] for _ in num_methods_range]
  times = [[] for _ in num_methods_range]
  sorting_times = []
  for _ in range(num_datasets):
    # Generate data.
    x = data_gen_fun(n)

    # Compute true median.
    median_index = int((n - 1) / 2)
    sort_start_time = time.time()
    x_sorted = np.sort(x)
    sort_end_time = time.time()
    sorting_times.append(sort_end_time - sort_start_time)
    median = x_sorted[median_index]
    # Numpy median implementation is the right median.  Left median is computed
    # above.  Doesn't really matter which is used though.
    # median = np.median(x)

    # Truncate to [min_value, max_value] range.
    x[x < min_value] = min_value
    x[x > max_value] = max_value

    estimates = []
    for method_num in num_methods_range:
      estimate_start_time = time.time()
      estimates.append(methods[method_num](
          x_sorted=x_sorted, num_estimates=num_trials_per_dataset))
      estimate_end_time = time.time()
      times[method_num].append(estimate_end_time - estimate_start_time)

    # Average over all trials to get a single number for each method on this
    # dataset.
    for method_num in num_methods_range:
      diffs[method_num].append(np.mean(np.abs(estimates[method_num] - median)))

  # Average over all datasets.
  avg_diffs = []
  std_of_diffs = []
  for method_num in num_methods_range:
    method_diffs = diffs[method_num]
    avg_diffs.append(np.mean(method_diffs))
    std_of_diffs.append(np.std(method_diffs))

  # Print stats.
  if verbose:
    print("Avg (L1) distance from target (true median):")
    for method_num in num_methods_range:
      print((
          "\t%s: %f +/- %f" %
          (names[method_num], avg_diffs[method_num], std_of_diffs[method_num])))

    # Average over all trials and datasets.
    print(("Avg sorting time (in milliseconds): %f\n" %
           (1000 * np.mean(sorting_times))))
    print("Avg estimate-generation time (in milliseconds):")
    for method_num in num_methods_range:
      print("\t%s: %f" % (names[method_num], 1000 * np.mean(times[method_num])))
    print()

  return (avg_diffs, std_of_diffs)


def main():
  random.seed(_RANDOM_SEED)

  # Methods to test.
  ss_method1 = functools.partial(
      private_medians.smooth_sensitivity_median, delta=_DELTA0)
  ss_method2 = functools.partial(
      private_medians.smooth_sensitivity_median, delta=_DELTA_SMALL)
  ss_method3 = functools.partial(
      private_medians.smooth_sensitivity_median, delta=_DELTA_LARGE)
  exp_method = private_medians.exponential_method_median
  method_arr = [ss_method1, ss_method2, ss_method3, exp_method]
  legend_names = [
      r"SmoothSens, $\delta={}$".format(_DELTA0),
      r"SmoothSens, $\delta={}$".format(_DELTA_SMALL),
      r"SmoothSens, $\delta={}$".format(_DELTA_LARGE), "Exp with $Duff$"
  ]
  assert len(method_arr) == len(legend_names)

  # Dataset generation functions, and reasonable bounds on min/max values.
  data_gen_funs = [
      lambda n: np.random.normal(0, 1, n), lambda n: np.random.uniform(0, 1, n),
      lambda n: np.random.beta(0.5, 0.5, n)
  ]
  data_gen_fun_names = ["Normal(0,1)", "Uniform(0,1)", "Beta(0.5,0.5)"]
  assert len(data_gen_funs) == len(data_gen_fun_names)
  min_values = [-10, 0, 0]
  max_values = [10, 1, 1]
  assert len(min_values) == len(max_values)
  assert len(data_gen_funs) == len(min_values)

  # Run tests with all dataset types.
  results = []
  for data_gen_fun_num in range(len(data_gen_funs)):
    data_gen_f = data_gen_funs[data_gen_fun_num]
    data_gen_fun_name = data_gen_fun_names[data_gen_fun_num]
    min_value = min_values[data_gen_fun_num]
    max_value = max_values[data_gen_fun_num]
    print("Data generated from " + data_gen_fun_name + "\n")

    truncated_methods = [
        functools.partial(method, min_value=min_value, max_value=max_value)
        for method in method_arr
    ]

    # Run tests with all values of epsilon.
    all_avg_diffs = []
    all_std_of_diffs = []
    for epsilon in _EPSILONS:
      if _VERBOSE:
        print("Epsilon = %f\n" % (epsilon))

      eps_methods = [
          functools.partial(method, epsilon=epsilon)
          for method in truncated_methods
      ]
      avg_diff_res, std_of_diff_res = run_medians_eval(
          data_gen_f, _DATASET_SIZE, min_value, max_value, _NUM_DATASETS,
          _NUM_TRIALS_PER_DATASET, eps_methods, legend_names, _VERBOSE)
      all_avg_diffs.append(avg_diff_res)
      all_std_of_diffs.append(std_of_diff_res)

    results.append((all_avg_diffs, all_std_of_diffs))

  # Make a plot for each dataset type, with a curve for each method.
  plt_styles = ["rs", "m>", "m<", "go"]
  assert len(plt_styles) == len(method_arr)
  num_epsilons = len(_EPSILONS)
  for data_gen_fun_num in range(len(data_gen_fun_names)):
    data_gen_fun_name = data_gen_fun_names[data_gen_fun_num]
    all_avg_diffs, all_std_of_diffs = results[data_gen_fun_num]
    for method_num in range(len(method_arr)):
      avg_diffs = []
      std_of_diffs = []
      for epsilon_num in range(num_epsilons):
        avg_diff = all_avg_diffs[epsilon_num][method_num]
        avg_diffs.append(avg_diff)
        std_of_diff = all_std_of_diffs[epsilon_num][method_num]
        std_of_diffs.append(std_of_diff)
      plt.errorbar(
          _EPSILONS,
          avg_diffs,
          yerr=std_of_diffs,
          fmt=plt_styles[method_num],
          label=legend_names[method_num],
          markersize=10)

    plt.yscale("log")
    plt.xlabel(r"$\epsilon$", fontsize=18, fontweight="bold")
    plt.ylabel(r"$|T(\mathcal{S}) - \mathcal{M}(\mathcal{S})|$", fontsize=20)
    plt.xlim(min(_EPSILONS) - 0.1, max(_EPSILONS) + 0.1)
    plt.title("Data from " + data_gen_fun_name)
    plt.gcf().set_size_inches(9, 7)
    plt.legend(prop={"size": 15}).get_frame().set_facecolor("w")

    if not os.path.exists("/tmp/private_medians"):
      os.makedirs("/tmp/private_medians")
    plt.savefig("/tmp/private_medians/" + data_gen_fun_name + ".pdf")
    plt.close()


if __name__ == "__main__":
  main()
